var qs          = require('querystring');
var url         = require('url');
var Entities    = require('html-entities').AllHtmlEntities;
var util        = require('./util');


var VIDEO_URL = 'https://www.youtube.com/watch?v=';
var getMetaItem = function(body, name) {
  return util.between(body, '<meta itemprop="${name}" content="', '">');
};


/**
 * Get video description from html
 *
 * @param {string} html
 * @return {string}
 */
exports.getVideoDescription = function(html) {
  var regex = /<p.*?id="eow-description".*?>(.+?)<\/p>[\n\r\s]*?<\/div>/im;
  var description = html.match(regex);
  return description ?
    Entities.decode(util.stripHTML(description[1])) : '';
};

/**
 * Get video media (extra information) from html
 *
 * @param {string} body
 * @return {Object}
 */
exports.getVideoMedia = function(body) {
  var mediainfo = util.between(body,
    '<div id="watch-description-extras">',
    '<div id="watch-discussion" class="branded-page-box yt-card">');
  if (mediainfo === '') {
    return {};
  }

  var regexp = /<h4 class="title">([\s\S]*?)<\/h4>[\s\S]*?<ul .*?class=".*?watch-info-tag-list">[\s\S]*?<li>([\s\S]*?)<\/li>(?:\s*?<li>([\s\S]*?)<\/li>)?/g;
  var contentRegexp = /(?: - (\d{4}) \()?<a .*?(?:href="([^"]+)")?.*?>(.*?)<\/a>/;
  var imgRegexp = /<img src="([^"]+)".*?>/;
  var media = {};

  var image = imgRegexp.exec(mediainfo);
  if (image) {
    media.image = url.resolve(VIDEO_URL, image[1]);
  }

  var match;
  while ((match = regexp.exec(mediainfo)) != null) {
	var key = match[1];
	var value = match[2];
	var detail = match[3];
    key = Entities.decode(key).trim().replace(/\s/g, '_').toLowerCase();
    var content = contentRegexp.exec(value);
    if (content) {
	  var year = content[1];
	  var mediaUrl = content[2];
	  var value2 = content[3];
      if (year) {
        media.year = parseInt(year);
      } else if (detail) {
        media.year = parseInt(detail);
      }
      value = value.slice(0, content.index);
      if (key !== 'game' || value2 !== 'YouTube Gaming') {
        value += value2;
      }
      media[key + '_url'] = url.resolve(VIDEO_URL, mediaUrl);
    }
    media[key] = Entities.decode(value);
  }
  return media;
};

/**
 * Get video Owner from html.
 *
 * @param {string} body
 * @return {Object}
 */
var userRegexp = /<a href="\/user\/([^"]+)/;
var verifiedRegexp = /<span .*?(aria-label="Verified")(.*?(?=<\/span>))/;
exports.getAuthor = function(body) {
  var ownerinfo = util.between(body,
    '<div id="watch7-user-header" class=" spf-link ">',
    '<div id="watch8-action-buttons" class="watch-action-buttons clearfix">');
  if (ownerinfo === '') {
    return {};
  }
  var channelName = Entities.decode(util.between(util.between(
    ownerinfo, '<div class="yt-user-info">', '</div>'), '>', '</a>'));
  var userMatch = ownerinfo.match(userRegexp);
  var verifiedMatch = ownerinfo.match(verifiedRegexp);
  var channelID = getMetaItem(body, 'channelId');
  var username = userMatch ? userMatch[1] : util.between(
    util.between(body, '<span itemprop="author"', '</span>'), '/user/', '">');
  return {
    id: channelID,
    name: channelName,
    avatar: url.resolve(VIDEO_URL, util.between(ownerinfo,
      'data-thumb="', '"')),
    verified: !!verifiedMatch,
    user: username,
    channel_url: 'https://www.youtube.com/channel/' + channelID,
    user_url: 'https://www.youtube.com/user/' + username,
  };
};


/**
 * Get video published at from html.
 *
 * @param {string} body
 * @return {string}
 */
exports.getPublished = function(body) {
  return Date.parse(getMetaItem(body, 'datePublished'));
};


/**
 * Get video published at from html.
 * Credits to https://github.com/paixaop.
 *
 * @param {string} body
 * @return {Array.<Object>}
 */
exports.getRelatedVideos = function(body) {
  var jsonStr = util.between(body, '\'RELATED_PLAYER_ARGS\': ', ',\n');
  var watchNextJson, rvsParams, secondaryResults;
  try {
    jsonStr = JSON.parse(jsonStr);
    watchNextJson = JSON.parse(jsonStr.watch_next_response);
    rvsParams = jsonStr.rvs.split(',').map(function(e) {return qs.parse(e)});
    secondaryResults = watchNextJson.contents.twoColumnWatchNextResults.secondaryResults.secondaryResults.results;
  }
  catch (err) {
    return [];
  }
  var videos = [];
  for (var result in secondaryResults) {
    var details = result.compactVideoRenderer;
    if (details) {
      try {
        var viewCount = details.viewCountText.simpleText;
        var shortViewCount = details.shortViewCountText.simpleText;
        var rvsDetails = rvsParams.find(function(elem) {return elem.id === details.videoId});
        if (!/^\d/.test(shortViewCount)) {
          shortViewCount = rvsDetails && rvsDetails.short_view_count_text || '';
        }
        viewCount = (/^\d/.test(viewCount) ? viewCount : shortViewCount).split(' ')[0];
        videos.push({
          id: details.videoId,
          title: details.title.simpleText,
          author: details.shortBylineText.runs[0].text,
          ucid: details.shortBylineText.runs[0].navigationEndpoint.browseEndpoint.browseId,
          author_thumbnail: details.channelThumbnail.thumbnails[0].url,
          short_view_count_text: shortViewCount.split(' ')[0],
          view_count: viewCount.replace(',', ''),
          length_seconds:  details.lengthText ?
            0 : 
            rvsParams && rvsParams.length_seconds + '',
          video_thumbnail: details.thumbnail.thumbnails[0].url,
        });
      } catch (err) {
        continue;
      }
    }
  }
  return videos;
};
